import logging
import os
import tempfile
import zipfile
from contextlib import contextmanager
from glob import glob
from typing import Optional
from uuid import uuid4

import yaml

from .logging import log

LOGGER = logging.getLogger(__name__)


@contextmanager
def ensure_directory_exists(directory_path: str) -> None:
    """
    Context manager and decorator to ensure a directory exists, and create it if it does not.
    :param directory_path: path to the folder
    :return: None
    """
    try:
        LOGGER.debug(f'Creating directory {directory_path}')
        if directory_path not in ['', '.']:
            os.makedirs(directory_path, exist_ok=True)
        yield
    finally:
        pass


@contextmanager
def ignore_exceptions(*exceptions: Exception) -> None:
    """
    Context manager and decorator to ignore Python exceptions. If no exception types are provided, all exceptions will
    be ignored. Should be used cautiously.
    :param exceptions: specific exceptions to be ignored
    :return: None
    """
    try:
        yield
    except tuple(exceptions) or Exception:
        pass


def safe_write(filepath: str, content: str) -> None:
    """
    Write content into a file (existing or not), and create the directory of the file if it does not exist.
    :param filepath: path to the file to be written
    :param content: content to be written in the file
    :return: None
    """
    with ensure_directory_exists(os.path.dirname(filepath)):
        with open(filepath, "w") as f:
            f.write(content)


def read_yaml(path: str) -> dict:
    """
    Read and load a YAML file safely.
    :param path: path to the YAML file to load.
    :return: loaded YAML file as a dict
    """
    return yaml.load(open(path, 'r'), Loader=yaml.SafeLoader)


def get_tmp_dirpath() -> str:
    """
    Generate temporary directory and return its path.
    :return: path to the temporary directory
    """
    dirpath = f'/tmp/{str(uuid4())}'
    with ensure_directory_exists(dirpath):
        return dirpath


def get_tmp_filepath(tmp_dirpath: Optional[str] = None, suffix: str = '') -> str:
    """
    Generate filepath for a file in a temporary folder, which can be specified or not.
    :param tmp_dirpath: optional directory, where the temporary file should be located in, when provided
    :param suffix: optional suffix to add to the filepath
    :return: generated filepath
    """
    filename_suffix = '_' + suffix if suffix else ''
    with tempfile.NamedTemporaryFile(suffix=filename_suffix) as tmp_file:
        tmp_dirpath = tmp_dirpath or os.path.dirname(tmp_file.name)
        suffix = os.path.basename(tmp_file.name)
    return os.path.join(tmp_dirpath, suffix)


def zip_files(input_filepath: str, output_filepath: str, exclude_filepath: str = '', if_exists: Optional[str] = 'fail'):
    """
    Function taking a path (or glob path pattern) and creating a ZIP file with all content matching this path.
    :param input_filepath: input file(s) pathname in the glob unix pathname pattern format. More information here:
        https://docs.python.org/3.7/library/glob.html#module-glob
    :param output_filepath: path to the ZIP file to be created
    :param exclude_filepath: file(s) pathname to be excluded from the ZIP file being created.
    :param if_exists: action to take if the output file already exists. Should be in ('fail', 'replace')
    :return: None
    """
    # If the output file already exists, handle it beforehand
    if os.path.exists(output_filepath):
        if if_exists == 'fail':
            raise FileExistsError(f'File {output_filepath} already exists. Set if_exists to "replace" to overwrite it.')
        elif if_exists == 'replace':
            log(f'File {output_filepath} is being replaced.', logger=LOGGER, level=logging.WARNING)
            os.remove(output_filepath)
        else:
            raise NotImplementedError
    # Create the output file
    input_files = glob(input_filepath, recursive=True)  # Need to be done outside of the with context manager
    excluded_files = glob(exclude_filepath, recursive=True)
    with zipfile.ZipFile(output_filepath, 'w', zipfile.ZIP_DEFLATED) as output:
        for filepath in (set(input_files) - set(excluded_files)):
            log(f'Writing: {filepath}', logger=LOGGER, level=logging.DEBUG)
            output.write(filepath)
    log(f'Successfully zipped "{input_filepath}" into "{output_filepath}"', logger=LOGGER, level=logging.INFO)


def unzip_file(input_filepath: str, output_dirpath: str = '.'):
    """
    Uncompress a ZIP file into a specific directory.
    :param input_filepath: path to the ZIP file to unzip
    :param output_dirpath: path to the directory to unzip the file into. Defaults to current working directory
    :return: None
    """
    file = zipfile.ZipFile(input_filepath, 'r', zipfile.ZIP_DEFLATED)
    file.extractall(path=output_dirpath)
    log(f'Successfully unzipped "{input_filepath}" into "{output_dirpath}"', logger=LOGGER, level=logging.INFO)



from functools import wraps

import pandas as pd

from .logging import log


def check_size_variation(logger=None):
    def decorator(func):
        @wraps(func)
        def wrapper(df, *args, **kwargs):
            nb_rows_input = df.shape[0]
            output = func(df, *args, **kwargs)
            nb_rows_output = output.shape[0]
            nb_rows_evolution = (nb_rows_output - nb_rows_input) / nb_rows_input * 100
            log(f'{func.__name__}: {nb_rows_input} -> {nb_rows_output} ({nb_rows_evolution:.1f}%)', logger)
            return output
        return wrapper
    return decorator


def configure_pandas(max_rows=None, max_columns=None, max_width=None, max_colwidth=None, logger=None):
    pd.set_option('display.max_rows', max_rows)
    pd.set_option('display.max_columns', max_columns)
    pd.set_option('display.max_colwidth', max_colwidth)
    pd.set_option('display.width', max_width)
    pd.set_option('display.expand_frame_repr', True)
    log('Pandas options were successfully set', logger=logger)



import logging
import os

from git import Repo
from git.exc import GitCommandError

from version import __version__
from dslib.utils.logging import configure_logging, deep_suppress_stdout_stderr
from dslib.wrappers.google.storage import StorageWrapper

# Variables about where to store the file
GSC_PROJECT = 'bbc-business-intelligence-data'
GSC_BUCKET = 'bbc-datascience-libraries'
GSC_DIRECTORY = 'python/'


# Set up logging
configure_logging()
LOGGER = logging.getLogger(__name__)

# Add version
repo = Repo()
try:
    LOGGER.info(f'Create git tag "{__version__}"')
    new_tag = repo.create_tag(__version__, message=f'Automatic tag "{__version__}"')
except GitCommandError as exception:
    raise ValueError(f'Invalid version "{__version__}": change it in dslib/__init__.py') from exception
LOGGER.info('Pushing local changes to remote git repository')
repo.remotes.origin.push(new_tag)

# Create source TAR.GZ package
local_dirpath = f'dist/'
LOGGER.info(f'Building package archive in {local_dirpath}')
with deep_suppress_stdout_stderr():
    os.system(f'python setup.py sdist --dist-dir {local_dirpath}')

# Check versions match
pkg_name = f'dslib-{__version__}.tar.gz'
pkg_local_path = os.path.join(local_dirpath, pkg_name)
assert pkg_name in os.listdir(local_dirpath)
LOGGER.info(f'Package {pkg_local_path} successfully created')

# Load it on Google Cloud Storage and make it public
storage = StorageWrapper.from_params(project=GSC_PROJECT, default_bucket=GSC_BUCKET)
pkg_distant_path = os.path.join(GSC_DIRECTORY, pkg_name)
storage.upload_file(pkg_local_path, pkg_distant_path, replace=False)  # Error if already exists
LOGGER.info(f'Public link to "{__version__}" package: {storage.get_public_url(pkg_distant_path)}')

# Load the "latest" version as well and make it public
pkg_name_latest = f'dslib-latest.tar.gz'
pkg_distant_path_latest = os.path.join(GSC_DIRECTORY, pkg_name_latest)
storage.upload_file(pkg_local_path, pkg_distant_path_latest, replace=True)
LOGGER.info(f'Public link to "latest" package: {storage.get_public_url(pkg_distant_path_latest)}')



import numpy as np


def predict_with_threshold(clf, df, threshold):
    """
    Perform class prediction with given trained binary classifier and threshold
    :param clf: binary classifier to be used
    :param df: input dataframe to predict on
    :param threshold: threshold to be used for classification
    :return: the classes predicted
    """
    # Assert threshold value is valid
    assert 0 < threshold < 1, 'Your threshold must be in interval ]0,1['
    # Use the model to predict probabilities
    y_prob = clf.predict_proba(df)[:, 1]
    # Choose a class based on the probabilities and the threshold
    return np.abs(np.round(y_prob - (threshold - 0.5)))  # Cut-off moved from 0.5 to threshold



from .caching import timed_cache
from .dataframe import check_size_variation, configure_pandas
from .io import ensure_directory_exists, ignore_exceptions, safe_write, read_yaml, get_tmp_dirpath, get_tmp_filepath, \
    zip_files, unzip_file
from .logging import configure_logging, log, timeit, suppress_stdout, deep_suppress_stdout_stderr
from .meta import resolve
from .ml import predict_with_threshold
from .templating import render_file, render_string
from .temporality import midpoint_date, date_range, change_timezone, force_timezone_onto_timestamp



import codecs
import os
import re

import setuptools

CURRENT_FILEPATH = os.path.abspath(os.path.dirname(__file__))
VERSION_FILENAME = 'version.py'


def find_version():
    # Source: https://packaging.python.org/guides/single-sourcing-package-version/
    with codecs.open(os.path.join(CURRENT_FILEPATH, VERSION_FILENAME), 'r') as f:
        version_file = f.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setuptools.setup(
    name="bbc-dslib",
    version=find_version(),
    author="Raphael Berly",
    author_email="raphael.berly@blablacar.com",
    description="A lib for the Data Science team at Blablacar",
    license='MIT',
    url="https://bitbucket.corp.blablacar.com/projects/lib/repos/dslib",
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires=['humanfriendly', 'pyyaml', 'jinja2', 'GitPython'],
    extras_require={
        'google': ['google-cloud-bigquery>=1.18.0', 'oauth2client', 'grpcio', 'pandas', 'pandas-gbq', 'tqdm',
                   'google-cloud-storage', 'google-api-python-client', 'pyarrow', 'google-cloud-bigquery-storage',
                   'gspread==3.6.0', 'gspread-dataframe==3.0.6'],
        'prophet': ['fbprophet'],
        'database': ['sqlalchemy', 'psycopg2-binary', 'PyMySQL', 'pandas'],
        'testing': ['pytest', 'pytest-cov', 'coverage', 'mock', 'testfixtures'],
    }
)
